﻿Type=Class
Version=6.8
ModulesStructureVersion=1
B4A=true
@EndOfDesignText@
'************************************************************************************
'
'This class module sets up voice recognition and is based on an excellent bit of
'investigation by stevel05 at:
'
'https://www.b4x.com/android/forum/threads/android-speech-recognition-api-wrapper.62959/
'
'Features:
'
'   o Full wrap of Android SpeechRecognizer API
'
'   o Supresses annoying beeps, conceptually as suggested by Rick Harris at:
'     https://www.b4x.com/android/forum/threads/chargeable-speechrecognizer-library.27409/page-3#post-220692
'
'   o Workaround for API's failure to raise ERROR_SPEECH_TIMEOUT onError event when 
'     speech timeout actually occurs
'
'   o Provides control not available thru Phone.VoiceRecognition (eg PREFER_OFFLINE)
'
'   o Comprehensive documentation
'
'Methods:
'
'   Initialize
'
'   Start
'
'   Flush
'
'   StartListening
'
'   StopListening
'
'Events:
'
'   Optional user specified callback on Android SpeechRecognizer events
'
'Properties:
'
'   IsRecognitionAvailable
'
'Requirements:
'
'   o B4A libraries:
'
'         JavaObject (version 2.05 or later)
'         Phone (version 2.40 or later)
'
'Update history:
'
'   19 Jan 17 - 1.0
'   24 Mar 17 - 2.0 (fixed new eccentricty - API has stopped issuing 
'                   ERROR_SPEECH_TIMEOUT onError event when speech timeout
'                   actually occurs)
'
'************************************************************************************
Private Sub Class_Globals
    
    Private SpeechRecognizerWrapper As JavaObject
    Private SpeechRecognizer As JavaObject
    Private SpeechRecognizerIntent As Intent
    Private SpeechRecognizer_Initialized As Boolean
    Private SpeechRecognizer_Started As Boolean
    Private SpeechRecognizer_Parent As Object
	Private SpeechRecognizer_Event_Name As String

	Private Obj_phone As Phone
	Private Obj_music_volume_restore_timer As Timer
	Private Obj_force_speech_timeout As Timer
	
	Private Gen_music_volume As Int
	Private Gen_speech_listening As Boolean
    
End Sub

'************************************************************************************
'
'This procedure gets control when Initialize method is called by parent module
'
'Input parameters are:
'
'        XSpeechRecognizer_Parent = pointer to parent module (module creating 
'                                   instance of this class)
'        XSpeechRecognizer_Event_Name = event name stub
'        XSpeechRecognizer_Beep_Mute_Time = amount of time music volume must be 
'                                           muted to eliminate annoying beep 
'                                           (millisecs) - this may need adjusting 
'                                           for different phones - 800 will 
'                                           typically eliminate all but a miniscule
'                                           beep every so often - to eliminate 
'                                           miniscule beep typically a value of 
'                                           4000-6000 or higher is required - set
'                                           to 0 to suppress this feature
'        XSpeechRecognizer_Speech_Timeout = amount of time between forced simulation
'                                           of ERROR_SPEECH_TIMEOUT onError events
'                                           (millisecs) - theoretically this is 5000
'                                           but may need to be marginally higher to 
'                                           avoid ERROR_RECOGNIZER_BUSY on_error
'                                           events - set to 0 to suppress this 
'                                           feature                                           
'
'Returns:
'
'       None
'
'Notes on this procedure:
'
'       o None
'
'************************************************************************************

'
'[Method] - initializes XSpeechRecognizer
'
'XSpeechRecognizer_Parent = pointer to parent module (module creating instance of this class)
'XSpeechRecognizer_Event_Name = event name stub
'XSpeechRecognizer_Beep_Mute_Time = amount of time music volume must be muted to eliminate annoying beep (millisecs) - this may need adjusting for different phones - 800 will typically eliminate all but a miniscule beep every so often - to eliminate miniscule beep typically a value of 4000-6000 or higher is required - set to 0 to suppress this feature
'XSpeechRecognizer_Speech_Timeout = amount of time between forced simulation of ERROR_SPEECH_TIMEOUT onError events (millisecs) - theoretically this is 5000 but may need to be marginally higher to avoid ERROR_RECOGNIZER_BUSY on_error events - set to 0 to suppress this feature 
Public Sub Initialize(XSpeechRecognizer_Parent As Object, XSpeechRecognizer_Event_Name As String, XSpeechRecognizer_Beep_Mute_Time As Int, XSpeechRecognizer_Speech_Timeout As Int)

    'Quit if already initialized
	If SpeechRecognizer_Initialized Then Return

	'Save input parameters that are used elsewhere in this class
	SpeechRecognizer_Parent = XSpeechRecognizer_Parent
    SpeechRecognizer_Event_Name = XSpeechRecognizer_Event_Name
    
    'Wrap API
	SpeechRecognizerWrapper.InitializeStatic("android.speech.SpeechRecognizer")
    
	'Create instance
	SpeechRecognizer = SpeechRecognizerWrapper.RunMethod("createSpeechRecognizer", Array(SpeechRecognizer.InitializeContext))

	'Set up music volume restoration timer
	If XSpeechRecognizer_Beep_Mute_Time > 0 Then Obj_music_volume_restore_timer.Initialize("Event_obj_music_restore_timer", XSpeechRecognizer_Beep_Mute_Time)
 
	'Set up ERROR_SPEECH_TIMEOUT onError simulator timer
	If XSpeechRecognizer_Speech_Timeout > 0 Then Obj_force_speech_timeout.Initialize("Event_obj_force_speech_timeout", XSpeechRecognizer_Speech_Timeout)
	
	'Flag object is initialized
	SpeechRecognizer_Initialized = True
    
End Sub

'************************************************************************************
'
'This procedure gets control when Start method is called by parent module
'
'Input parameters are:
'
'       None
'
'Returns:
'
'       None
'
'Notes on this procedure:
'
'       o This has been configured for voice activation with keywords preferably 
'         using an offline speech recognition engine - if your need is different 
'         (web search, using remote speech recognition engine etc) then you will have 
'         to play around with all SpeechRecognizerIntent... statements
'       o SpeechRecognizerIntent constants are documented in bottom half of:
'         https://developer.android.com/reference/android/speech/RecognizerIntent.html
'
'************************************************************************************

'
'[Method] - starts XSpeechRecognizer
Public Sub Start

    'Quit if not initialized
    If Not(SpeechRecognizer_Initialized) Then Return
    
    'Quit if already started
    If SpeechRecognizer_Started Then Return
    
    'Quit if speech recognition not available
	If Not(IsRecognitionAvailable) Then Return
	
	'Get current music volume if it is not in process of being restored
	If Not(Obj_music_volume_restore_timer.Enabled) Then Gen_music_volume = Obj_phone.GetVolume(Obj_phone.VOLUME_MUSIC)

	'Comments in this area lifted straight from:
	'https://developer.android.com/reference/android/speech/RecognizerIntent.html
	
	'Starts an activity that will prompt the user for speech without requiring the
    'user's visual attention or touch input. It will send it through a speech 
    'recognizer, and either synthesize speech for a web search result or trigger 
    'another type of action based on the user's speech. This activity may be launched 
    'while device is locked in a secure mode. Special care must be taken to ensure 
    'that the voice actions that are performed while hands free cannot compromise the 
    'device's security. The activity should check the value of the EXTRA_SECURE extra 
    'to determine whether the device has been securely locked. If so, the activity 
    'should either restrict the set of voice actions that are permitted or require 
    'some form of secure authentication before proceeding. To ensure that the 
    'activity's user interface is visible while the lock screen is showing, the 
    'activity should set the FLAG_SHOW_WHEN_LOCKED window flag. Otherwise the 
    'activity's user interface may be hidden by the lock screen. The activity should 
    'take care not to leak private information when the device is securely locked.
    SpeechRecognizerIntent.Initialize("android.speech.action.VOICE_SEARCH_HANDS_FREE", Null)
	
    'SpeechRecognizerIntent.PutExtra("android.speech.extra.LANGUAGE_PREFERENCE","en")
    
    'The extra key used in an intent to the speech recognizer for voice search. Not
    'generally to be used by developers. The system search dialog uses this, for 
    'example, to set a calling package for identification by a voice search API. If 
    'this extra is set by anyone but the system process, it should be overridden by 
    'the voice search implementation.
    SpeechRecognizerIntent.PutExtra("calling_package", Application.PackageName)

    'SpeechRecognizerIntent.PutExtra("android.speech.extra.LANGUAGE_MODEL","free_form")

    'Optional limit on the maximum number of results to return. If omitted the 
    'recognizer will choose how many results to return. Must be an integer.
    SpeechRecognizerIntent.PutExtra("android.speech.extra.MAX_RESULTS", 1)
    
    'Optional boolean, to be used with ACTION_RECOGNIZE_SPEECH,
    'ACTION_VOICE_SEARCH_HANDS_FREE, ACTION_WEB_SEARCH to indicate whether to only 
    'use an offline speech recognition engine. The default is false, meaning that 
    'either network or offline recognition engines may be used.
    SpeechRecognizerIntent.PutExtra("android.speech.extra.PREFER_OFFLINE", True)
    
    'Optional boolean to indicate whether partial results should be returned by the 
    'recognizer as the user speaks (default is false). The server may ignore a 
    'request for partial results in some or all cases.
    SpeechRecognizerIntent.PutExtra("android.speech.extra.PARTIAL_RESULTS", True)
    
    'The amount of time that it should take after we stop hearing speech to consider 
    'the input complete. Note that it is extremely rare you'd want to specify this 
    'value in an intent. If you don't have a very good reason to change these, you 
    'should leave them as they are. Note also that certain values may cause undesired 
    'or unexpected results - use judiciously! Additionally, depending on the 
    'recognizer implementation, these values may have no effect.
    'SpeechRecognizerIntent.PutExtra("android.speech.extras.SPEECH_INPUT_COMPLETE_SILENCE_LENGTH_MILLIS", 100)
    
    'The minimum length of an utterance. We will not stop recording before this 
    'amount of time. Note that it is extremely rare you'd want to specify this value 
    'in an intent. If you don't have a very good reason to change these, you should 
    'leave them as they are. Note also that certain values may cause undesired or 
    'unexpected results - use judiciously! Additionally, depending on the recognizer 
    'implementation, these values may have no effect.
    'SpeechRecognizerIntent.PutExtra("android.speech.extras.SPEECH_INPUT_MINIMUM_LENGTH_MILLIS", 100)
    
    'The amount of time that it should take after we stop hearing speech to consider 
    'the input possibly complete. This is used to prevent the endpointer cutting off 
    'during very short mid-speech pauses. Note that it is extremely rare you'd want 
    'to specify this value in an intent. If you don't have a very good reason to 
    'change these, you should leave them as they are. Note also that certain values 
    'may cause undesired or unexpected results - use judiciously! Additionally, 
    'depending on the recognizer implementation, these values may have no effect.
    'SpeechRecognizerIntent.PutExtra("android.speech.extras.SPEECH_INPUT_POSSIBLY_COMPLETE_SILENCE_LENGTH_MILLIS", 100)

    'Set up to receive events via SpeechRecognizer_Event
    Private Event As Object = SpeechRecognizer.CreateEvent("android.speech.RecognitionListener", "SpeechRecognizer" ,"")
    SpeechRecognizer.RunMethod("setRecognitionListener", Array(Event))

    'Flag object is started
    SpeechRecognizer_Started = True

End Sub

'************************************************************************************
'
'This procedure gets control when IsRecognitionAvailable property is got by parent 
'module
'
'Input parameters are:
'
'       None
'
'Returns:
'
'       Flag indicating whether voice recognition is available or not
'
'Notes on this procedure:
'
'       o None
'
'************************************************************************************

'
'[Property] - gets XSpeechRecognizer IsRecognitionAvailable
Public Sub IsRecognitionAvailable As Boolean

    'Quit if not initialized
    If Not(SpeechRecognizer_Initialized) Then Return False

    Private wrk_jo As JavaObject
    wrk_jo.InitializeContext
    Return SpeechRecognizer.RunMethod("isRecognitionAvailable", Array(wrk_jo))
    
End Sub

'************************************************************************************
'
'This procedure gets control when StartListening method is called by parent module
'
'Input parameters are:
'
'       None
'
'Returns:
'
'       None
'
'Notes on this procedure:
'
'       o None
'
'************************************************************************************

'
'[Method] - starts XSpeechRecognizer listening
Public Sub StartListening

    'Quit if not started
	If Not(SpeechRecognizer_Started) Then Return
 	
	'Quit if already listening
	If Gen_speech_listening Then Return

	'Suppress music volume
	Music_suppress
	
	Try
    
		SpeechRecognizer.RunMethod("startListening", Array(SpeechRecognizerIntent))
	
	Catch
		
	End Try
	
	'Flag listening
	Gen_speech_listening = True
	
	'If simulating ERROR_SPEECH_TIMEOUT onError events...
	If Obj_force_speech_timeout.IsInitialized Then
	
		'Toggle ERROR_SPEECH_TIMEOUT onError simulator timer off/on to reset timer 
		'countdown
		Obj_force_speech_timeout.Enabled = False
		Obj_force_speech_timeout.Enabled = True
		
	End If
   
End Sub

'************************************************************************************
'
'This procedure gets control when StopListening method is called by parent module
'
'Input parameters are:
'
'       None
'
'Returns:
'
'       None
'
'Notes on this procedure:
'
'       o None
'
'************************************************************************************

'
'[Method] - stops XSpeechRecognizer listening
Public Sub StopListening

    'Quit if not started
	If Not(SpeechRecognizer_Started) Then Return
	
	'Quit if already not listening
	If Not(Gen_speech_listening) Then Return

	'Suppress music volume
	Music_suppress
	
	Try
    
		SpeechRecognizer.RunMethod("stopListening", Null)
	
	Catch
		
	End Try

	'Disable ERROR_SPEECH_TIMEOUT onError simulator timer
	Obj_force_speech_timeout.Enabled = False
		
	'Flag not listening
	Gen_speech_listening = False
		
End Sub

'************************************************************************************
'
'This procedure gets control when Flush method is called by parent module
'
'Input parameters are:
'
'       None
'
'Returns:
'
'       None
'
'Notes on this procedure:
'
'       o None
'
'************************************************************************************

'
'[Method] - flushes XSpeechRecognizer
Public Sub Flush

    'Quit if not started
    If Not(SpeechRecognizer_Started) Then Return
	
	'Stop listening
	StopListening
	
	Try
    
		'Sink it
		SpeechRecognizer.RunMethod("destroy", Null)
	
	Catch
		
	End Try

	'Reset music volume immediately
	If Obj_music_volume_restore_timer.IsInitialized Then Event_obj_music_restore_timer_Tick

	'Disable ERROR_SPEECH_TIMEOUT onError simulator timer
	Obj_force_speech_timeout.Enabled = False
   
    'Flag object is stopped
    SpeechRecognizer_Started = False
    
End Sub

'************************************************************************************
'
'This procedure gets control when event is raised by Android SpeechRecognizer, as
'defined in Start method
'
'Input parameters are:
'
'       Event_name = Android SpeechRecognizer event name
'       Args() = arguments
'
'Returns:
'
'       None
'
'Notes on this procedure:
'
'       o None
'
'************************************************************************************
Private Sub SpeechRecognizer_Event(Event_name As String, Args() As Object)

	'Suppress music volume
	If Event_name <> "onRmsChanged" Then Music_suppress
    
    'Branch based on event name...
    Select Event_name
        
		Case "onBeginningOfSpeech"

			'If parent module has appropriate event procedure...
            If SubExists(SpeechRecognizer_Parent, SpeechRecognizer_Event_Name & "_onBeginningOfSpeech") Then

                'Fire it
				CallSubDelayed(SpeechRecognizer_Parent, SpeechRecognizer_Event_Name & "_onBeginningOfSpeech")

            End If
            
		Case "onEndOfSpeech"

            'If parent module has appropriate event procedure...
            If SubExists(SpeechRecognizer_Parent, SpeechRecognizer_Event_Name & "_onEndOfSpeech") Then

                'Fire it
				CallSubDelayed(SpeechRecognizer_Parent, SpeechRecognizer_Event_Name & "_onEndOfSpeech")

            End If
            
		Case "onError"
			
			'An error automatically stops listening
			Gen_speech_listening = False

			'Do not use Flush here because it resets music volume immediately which can
			'cause a beep
	
			Try
    
				'Sink it
				SpeechRecognizer.RunMethod("destroy", Null)
	
			Catch
		
			End Try
 
			'Flag object is stopped
			SpeechRecognizer_Started = False
			
			'Restart
			Start

			'If parent module has appropriate event procedure...
			If SubExists(SpeechRecognizer_Parent, SpeechRecognizer_Event_Name & "_onError") Then
             
				Private wrk_errno As Int = Args(0)

				'Fire it
				CallSubDelayed3(SpeechRecognizer_Parent, SpeechRecognizer_Event_Name & "_onError", wrk_errno, Get_error_msg(wrk_errno))

			End If
           
		Case "onResults"
			
			'A result automatically stops listening
			Gen_speech_listening = False
            
            'If parent module has appropriate event procedure...
			If SubExists(SpeechRecognizer_Parent, SpeechRecognizer_Event_Name & "_onResults") Then

				Private wrk_results As JavaObject = Args(0)
            
				Private wrk_words As List = wrk_results.RunMethod("getStringArrayList", Array("results_recognition"))

                'Fire it
				CallSubDelayed2(SpeechRecognizer_Parent, SpeechRecognizer_Event_Name & "_onResults", wrk_words)

			End If
            
        Case "onRmsChanged"
            
            'If parent module has appropriate event procedure...
			If SubExists(SpeechRecognizer_Parent, SpeechRecognizer_Event_Name & "_onRmsChanged") Then
            
				Private wrk_RmsDB As Float = Args(0)

                'Fire it
				CallSubDelayed2(SpeechRecognizer_Parent, SpeechRecognizer_Event_Name & "_onRmsChanged", wrk_RmsDB)

            End If
            
	End Select
    
End Sub

'************************************************************************************
'
'This procedure mutes music volume to suppress annoying beeps and then initiates its
'restoration 
'
'Input parameters are:
'
'       None
'
'Returns:
'
'       None
'
'Notes on this procedure:
'
'       o None
'
'************************************************************************************
Private Sub Music_suppress

	'Quit if music volume not being suppressed
	If Not(Obj_music_volume_restore_timer.IsInitialized) Then Return
	
	'Get current music volume if it is not in process of being restored
	If Not(Obj_music_volume_restore_timer.Enabled) Then	Gen_music_volume = Obj_phone.GetVolume(Obj_phone.VOLUME_MUSIC)

	'Suppress music volume
	Obj_phone.SetVolume(Obj_phone.VOLUME_MUSIC, 0, False)

	'Toggle music restoration timer off/on to reset timer countdown
	Obj_music_volume_restore_timer.Enabled = False
	Obj_music_volume_restore_timer.Enabled = True
	
End Sub

'************************************************************************************
'
'This procedure gets control when Obj_music_restore_timer Timer object times out
'
'Input parameters are:
'
'       None
'
'Returns:
'
'       None
'
'Notes on this procedure:
'
'       o None
'
'************************************************************************************
Private Sub Event_obj_music_restore_timer_Tick

	'Disable this timer
	Obj_music_volume_restore_timer.Enabled = False
	
	'Restore music volume
	Obj_phone.SetVolume(Obj_phone.VOLUME_MUSIC, Gen_music_volume, False)
	
End Sub

'************************************************************************************
'
'This procedure gets control when Obj_force_speech_timeout Timer object times out
'
'Input parameters are:
'
'       None
'
'Returns:
'
'       None
'
'Notes on this procedure:
'
'       o None
'
'************************************************************************************
Private Sub Event_obj_force_speech_timeout_Tick
	
	'Simulate ERROR_SPEECH_TIMEOUT onError event
	SpeechRecognizer_Event("onError", Array(SpeechRecognizerWrapper.GetField("ERROR_SPEECH_TIMEOUT")))
	
End Sub

'************************************************************************************
'
'This procedure converts an Android SpeechRecognizer error number to an error message
'
'Input parameters are:
'
'       Error_code = error code as supplied by onError
'
'Returns:
'
'       Error message equivalent of Error_code
'
'Notes on this procedure:
'
'       o None
'
'************************************************************************************
Private Sub Get_error_msg(Error_code As Int) As String
    
    Select Error_code
        
        Case SpeechRecognizerWrapper.GetField("ERROR_AUDIO")
            
            Return "Audio error"
            
         Case SpeechRecognizerWrapper.GetField("ERROR_CLIENT")
             
             Return "Client error"
            
         Case SpeechRecognizerWrapper.GetField("ERROR_INSUFFICIENT_PERMISSIONS")
            
            Return "Insufficient permissions"
            
         Case SpeechRecognizerWrapper.GetField("ERROR_NETWORK")
            
             Return "Network error"
            
         Case SpeechRecognizerWrapper.GetField("ERROR_NETWORK_TIMEOUT")
            
             Return "Network timeout"
            
          Case SpeechRecognizerWrapper.GetField("ERROR_NO_MATCH")
             
             Return "No match"
            
          Case SpeechRecognizerWrapper.GetField("ERROR_RECOGNIZER_BUSY")
            
             Return "Recognizer busy"
            
          Case SpeechRecognizerWrapper.GetField("ERROR_SERVER")
             
             Return "Server error"
            
         Case SpeechRecognizerWrapper.GetField("ERROR_SPEECH_TIMEOUT")
            
            Return "Speech timeout"
            
         Case Else
            
             Return "Unknown error (code " & Error_code & ")"
            
    End Select
    
End Sub
